==> opteryx-0.17.0a761-cp39-cp39-win_amd64.whl: opteryx/planner/cost_based_optimizer/strategies/distinct_pushdown.py <==
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Distinct Pushdown

Type: Heuristic
Goal: Reduce Rows

Rules:
    - DISTINCT ON can't get pushed
"""

from orso.tools import random_string

from opteryx.managers.expression import NodeType
from opteryx.managers.expression import get_all_nodes_of_type
from opteryx.models import LogicalColumn
from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext

PASSABLE_AGGREGATIONS = ("MIN", "MAX")
"""
Aggregations we can push the DISTINCT past
"""


class DistinctPushdownStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        """ """
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if (node.node_type == LogicalPlanStepType.Distinct) and node.on is None:
            node.nid = context.node_id
            node.plan_path = context.optimized_plan.trace_to_root(context.node_id)
            context.collected_distincts.append(node)
            context.optimized_plan.remove_node(context.node_id, heal=True)
            return context

        if (
            node.node_type == LogicalPlanStepType.Join
            and context.collected_distincts
            and node.type == "cross join"
            and node.unnest_target is not None
        ):
            node.distinct = True
            context.optimized_plan[context.node_id] = node
            context.collected_distincts.clear()
            return context

        if node.node_type in (
            LogicalPlanStepType.Join,
            LogicalPlanStepType.Scan,
            LogicalPlanStepType.AggregateAndGroup,
            LogicalPlanStepType.Aggregate,
        ):
            # anything we couldn't push, we need to put back
            for distinct in context.collected_distincts:
                for nid in distinct.plan_path:
                    if nid in context.optimized_plan:
                        context.optimized_plan.insert_node_before(distinct.nid, distinct, nid)
                        break
            return context

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan

==> opteryx-0.17.2a816-cp39-cp39-win_amd64.whl: opteryx/planner/cost_based_optimizer/strategies/distinct_pushdown.py <==
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Optimization Rule - Distinct Pushdown

Type: Heuristic
Goal: Reduce Rows

This is a very specific rule, on a CROSS JOIN UNNEST, if the result
is the only column in a DISTINCT clause, we push the DISTINCT into
the JOIN.

We've written as a Optimization rule rather than in the JOIN code
as it is expected other instances of pushing DISTINCT may be found.

Order:
    This plan must run after the Projection Pushdown
"""

from opteryx.planner.logical_planner import LogicalPlan
from opteryx.planner.logical_planner import LogicalPlanNode
from opteryx.planner.logical_planner import LogicalPlanStepType

from .optimization_strategy import OptimizationStrategy
from .optimization_strategy import OptimizerContext

"""
Aggregations we can push the DISTINCT past
"""


class DistinctPushdownStrategy(OptimizationStrategy):
    def visit(self, node: LogicalPlanNode, context: OptimizerContext) -> OptimizerContext:
        if not context.optimized_plan:
            context.optimized_plan = context.pre_optimized_tree.copy()  # type: ignore

        if (node.node_type == LogicalPlanStepType.Distinct) and node.on is None:
            node.nid = context.node_id
            context.collected_distincts.append(node)
            return context

        if (
            node.node_type == LogicalPlanStepType.Join
            and context.collected_distincts
            and node.type == "cross join"
            and node.unnest_target is not None
            and node.pre_update_columns == {node.unnest_target.identity}
        ):
            # Very specifically testing for a DISTINCT on the unnested column, only.
            # In this situation we do the DISTINCT on the intermediate results of the CJU,
            # this means we create smaller tables out of the CROSS JOIN => faster
            node.distinct = True
            context.optimized_plan[context.node_id] = node
            for distict_node in context.collected_distincts:
                context.optimized_plan.remove_node(distict_node.nid, heal=True)
            context.collected_distincts.clear()
            return context

        if node.node_type in (
            LogicalPlanStepType.Join,
            LogicalPlanStepType.Scan,
            LogicalPlanStepType.AggregateAndGroup,
            LogicalPlanStepType.Aggregate,
            LogicalPlanStepType.Subquery,
        ):
            # we don't push past here
            context.collected_distincts.clear()

        return context

    def complete(self, plan: LogicalPlan, context: OptimizerContext) -> LogicalPlan:
        # No finalization needed for this strategy
        return plan
